"""
AWS Request Signater Version 4
@auther: Vubon Roy
"""
import hashlib
import hmac

_AWS_REQUEST = "aws4_request"
_AWS4 = "AWS4"
_AWS_ALGORITHM = "AWS4-HMAC-SHA256"
_DEFAULT_CONTENT_TYPE = "application/x-amz-json-1.0"


class SignV4:
    """
    AWS Request Signature Version 4
    """

    def __init__(self, aws_access_key: str, aws_secret_key: str, aws_region: str, aws_service: str, date_time):
        self._signed_headers = 'content-type;host;x-amz-date'
        self._aws_access_key = aws_access_key
        self._aws_secret_key = aws_secret_key
        self.aws_region = aws_region
        self.aws_service = aws_service
        self.date_stamp = date_time.strftime('%Y%m%d')
        self.amz_date = date_time.strftime('%Y%m%dT%H%M%SZ')

    @staticmethod
    def hash_sha256(message: str) -> str:
        """
        :param message:
        :return:
        """
        return hashlib.sha256(message.encode('utf-8')).hexdigest()

    @staticmethod
    def sign(key: bytes, message: str) -> bytes:
        """
        Signing message with key
        :param key: bytes: Key as bytes
        :param message: str: Client message
        :return: return as bytes
        :rtype: bytes
        """
        return hmac.new(key=key, msg=message.encode('utf-8'), digestmod=hashlib.sha256).digest()

    def get_signature_key(self) -> bytes:
        """
        :return:
        """
        return self.sign(
            key=self.sign(
                key=self.sign(
                    key=self.sign(
                        key=(_AWS4 + self._aws_secret_key).encode('utf-8'), message=self.date_stamp
                    ),
                    message=self.aws_region
                ),
                message=self.aws_service
            ),
            message=_AWS_REQUEST
        )

    def get_canonical_headers(self, headers: dict) -> list:
        """
        :param headers:
        :return:
        """
        content_type = headers.get("content-type", _DEFAULT_CONTENT_TYPE)
        host = headers.get("host")
        return ['content-type:' + content_type + '\n' + 'host:' + host + '\n' + 'x-amz-date:' + self.amz_date + '\n']

    def get_canonical_request(self, **kwargs: dict) -> str:
        """
        :param kwargs:
         canonical = {
            "headers": {
                "content-type": "application/json",
                "host": "hello.com",
                "date": "datetime object UTC format"
            },
            "method": "POST",
            "payload": {"content": "hello world"},
            "canonical_uri": "hello/test",
            "canonical_querystring": "qos=1"
        }
        :return:
        """
        headers = kwargs.pop("headers")
        payload = kwargs.pop("payload")
        payload_hash = [self.hash_sha256(payload)]
        canonical_headers = self.get_canonical_headers(headers)

        _request = []
        for key, value in kwargs.items():
            _request.append(value)

        canonical_request = _request + canonical_headers + [self._signed_headers] + payload_hash
        return "\n".join([str(item) for item in canonical_request])

    def get_credential_scope(self) -> str:
        """
        :return:
        """
        return self.date_stamp + "/" + self.aws_region + "/" + self.aws_service + "/" + _AWS_REQUEST

    def string_to_sign(self, **kwargs) -> str:
        """
        :param kwargs:
        :return:
        """
        credential_scope = self.get_credential_scope()
        canonical_request = self.get_canonical_request(**kwargs)
        return _AWS_ALGORITHM + "\n" + self.amz_date + "\n" + credential_scope + "\n" + self.hash_sha256(
            canonical_request)

    def signature(self, **kwargs):
        """
        :return:
        """
        return hmac.new(self.get_signature_key(), self.string_to_sign(**kwargs).encode('utf-8'),
                        hashlib.sha256).hexdigest()

    def authorization_header(self, **kwargs) -> dict:
        """
        :return:
        """
        credential_scope = self.get_credential_scope()
        first_part = _AWS_ALGORITHM + " " + "Credential=" + self._aws_access_key + "/" + credential_scope
        second_part = ', SignedHeaders=' + self._signed_headers + ", " + 'Signature=' + self.signature(**kwargs)
        return {
            'Content-Type': "application/x-amz-json-1.0",
            'X-Amz-Date': self.amz_date,
            'Authorization': first_part + second_part,
        }



"""
Testing
"""
import os
import datetime

import unittest

from AWSSignV4.client import Client, get_host


class ClientTest(unittest.TestCase):

    def setUp(self) -> None:
        """Setup all initiale data"""
        self.client = Client(
            aws_region='<AWS Region>',
            aws_service='iotdevicegateway',
            aws_access_key=os.environ.get("aws_access_key"),
            aws_secret_key=os.environ.get("aws_secret_key"),
            date_time=datetime.datetime.now().utcnow()
        )
        self.endpoint = "<Your AWS IoT Core Endpoint>"
        self.client.build_canonical(canonical_uri="/topics/hello", canonical_querystring="qos=1")

    def test_post(self) -> None:
        payload = {"iot": "core"}
        res = self.client.post(self.endpoint, data=payload)
        self.assertEqual(res.status_code, 200)

    def test_get_host(self):
        res = get_host(self.endpoint)
        self.assertEqual(res, "a2qacovkwccbxk-ats.iot.us-east-1.amazonaws.com")



"""AWS Signature Version 4 Cleint"""
import json
from urllib.parse import urlparse

import requests
from AWSSignV4.signv4 import SignV4


def get_host(url: str) -> str:
    """
    :param url:
    :return:
    """
    parse = urlparse(url)
    return parse.hostname


class Client(SignV4):
    """
    Cleint layer
    """

    def __init__(self, aws_access_key: str, aws_secret_key: str, aws_region: str, aws_service: str, date_time):
        super(Client, self).__init__(aws_access_key, aws_secret_key, aws_region, aws_service, date_time)
        self.canonical_uri = None
        self.canonical_querystring = None

    def build_canonical(self, canonical_uri: str, canonical_querystring: str):
        """
         # "canonical_uri": "/topics/hello",
        # "canonical_querystring": "qos=1"
        :return:
        """

        self.canonical_querystring = canonical_querystring
        self.canonical_uri = canonical_uri

    def post(self, url: str, data):
        """
        :return:
        """
        payload = json.dumps(data)
        url += self.canonical_uri + ("?" + self.canonical_querystring) if self.canonical_querystring else None
        canonical = {
            "headers": {
                "host": get_host(url)
            },
            "method": "POST",
            "payload": payload,
        }
        canonical.update(
            {"canonical_uri": self.canonical_uri, "canonical_querystring": self.canonical_querystring})
        return requests.post(url, data=payload, headers=self.authorization_header(**canonical))



"""
AWS Request Signature4
"""
import os
import re

from setuptools import setup, find_packages

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


def get_version(package: str) -> str:
    """
    :param package:
    :return:
    """
    with open(os.path.join(BASE_DIR, f'{package}/__version__.py')) as version:
        version = version.readlines()[-1]
    match = re.search("__version__ = ['\"]([^'\"]+)['\"]", version)
    assert match is not None
    return match.group(1)


def get_log_description():
    """
    :return:
    """
    with open('README.md') as readme:
        with open('CHANGELOG.md') as changelog:
            return readme.read() + "\n\n" + changelog.read()


setup(
    name='AWSSignV4',
    version=get_version('AWSSignV4'),
    author='Vubon Roy',
    author_email='vubon.roy@gmail.com',
    description='This package will help to use AWS service without SDK',
    url='https://github.com/vubon/AWSSignV4',
    project_urls={
        "Documentation": "https://github.com/vubon/AWSSignV4/blob/master/docs/GUIDE.md"
    },
    packages=find_packages(),
    long_description=get_log_description(),
    long_description_content_type="text/markdown",
    license='MIT',
    platforms='Python',
    install_requires=[
        'requests'
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)
